import { makeAutoObservable, runInAction } from "mobx";
import api, { CredentialRecordDto } from "@dashboard/engine-api";
import { toast } from "react-toastify";
import displayError from "@dashboard/utils/displayError";

class CredentialListStore {
  public loading = false;
  public items: CredentialRecordDto[] = [];

  public statusChangeId: string | null = null;
  constructor() {
    makeAutoObservable(this);
  }

  public async loadAll(connectionId: string | null) {
    runInAction(() => (this.loading = true));
    try {
      const items = await api.fetchCredentials(
        undefined,
        connectionId || undefined,
      );
      runInAction(() => {
        this.items = items;
        this.loading = false;
      });
    } catch (e: unknown) {
      displayError(e);
      runInAction(() => (this.loading = false));
    }
  }

  public async accept(credId: string) {
    runInAction(() => {
      this.statusChangeId = credId;
    });
    try {
      const credRecord = await api.acceptCredential({
        credentialId: credId,
        credentialUrl: "",
      });
      runInAction(() => {
        const index = this.items.findIndex((p) => p.id === credRecord.id);
        if (index >= 0) {
          this.items[index] = credRecord;
        }
        this.statusChangeId = null;
      });
      toast.success("Credential offer accepted");
    } catch (e: unknown) {
      displayError(e);
      runInAction(() => {
        this.loading = false;
        this.statusChangeId = null;
      });
    }
  }

  public async decline(credId: string) {
    runInAction(() => {
      this.statusChangeId = credId;
    });
    try {
      const credRecord = await api.declineCredential(credId);
      runInAction(() => {
        const index = this.items.findIndex((p) => p.id === credRecord.id);
        if (index >= 0) {
          this.items[index] = credRecord;
        }
        this.statusChangeId = null;
      });
      toast.success("Credential offer declined");
    } catch (e: unknown) {
      displayError(e);
      runInAction(() => {
        this.loading = false;
        this.statusChangeId = null;
      });
    }
  }

  public async remove(credId: string) {
    runInAction(() => {
      this.statusChangeId = credId;
    });
    try {
      await api.deleteCredentialById(credId);
      runInAction(() => {
        const index = this.items.findIndex((p) => p.id === credId);
        if (index >= 0) {
          this.items.splice(index, 1);
        }
        this.statusChangeId = null;
      });
      toast.success("Credential record deleted successfully");
    } catch (e: unknown) {
      displayError(e);
      runInAction(() => {
        this.loading = false;
        this.statusChangeId = null;
      });
    }
  }
}

export type { CredentialListStore };

export default CredentialListStore;
