import React, { useEffect, useState } from "react";
import { observer } from "mobx-react";
import { Form, Input, Button, Spin, Select } from "antd";
import Modal, { FcProps } from "@dashboard/components/Modal";
import { CredentialOfferResponseDto } from "@dashboard/engine-api";
import OfferCredentialDialogStore from "./OfferCredentialDialogStore";

export interface OfferCredentialDialogProps {
  connectionId?: string;
  onCredentialOffered: (cred: CredentialOfferResponseDto) => void;
}

const OfferCredentialDialog = observer(
  ({ setTitle, onClose, data }: FcProps<OfferCredentialDialogProps>) => {
    const [store] = useState(() => new OfferCredentialDialogStore());
    useEffect(() => {
      if (data.connectionId) {
        setTitle("Offer credential to " + data.connectionId);
      } else {
        setTitle("Offer connectionless credential");
      }
      store.init(data.connectionId);
    }, [store, setTitle, data]);

    const onFinish = async ({
      attributes,
    }: {
      attributes: { name: string; value: string }[];
    }) => {
      const response = await store.offerCredential(attributes);
      if (response) {
        data.onCredentialOffered(response);
        onClose();
      }
    };

    const attributesInitialValue =
      store.schema?.attributes.map((p) => ({
        name: p,
        value: "",
      })) || [];

    return (
      <Modal.Body>
        {store.initLoading && <Spin>Loading credential definitions...</Spin>}
        {store.schemaLoading && (
          <Spin>Loading credential definition attributes...</Spin>
        )}

        <div>
          <Form
            labelCol={{ span: 8 }}
            wrapperCol={{ span: 16 }}
            style={{ maxWidth: 600 }}
            autoComplete="off"
          >
            <Form.Item
              label="Credential definition"
              name="credDef"
              rules={[
                {
                  required: true,
                  message: "Please select credential definition!",
                },
              ]}
            >
              <Select
                style={{ width: 300 }}
                value={store.selectedCredDefId}
                showSearch
                onChange={(val) => store.selectCredDef(val)}
                options={store.credDefList.map((p) => ({
                  value: p.id,
                  label: p.tag || p.id,
                }))}
              />
            </Form.Item>
          </Form>
          {}
          {store.schema && (
            <Form
              labelCol={{ span: 8 }}
              wrapperCol={{ span: 16 }}
              style={{ maxWidth: 600 }}
              initialValues={{}}
              onFinish={onFinish}
              autoComplete="off"
            >
              <Form.List
                name="attributes"
                initialValue={attributesInitialValue}
              >
                {(fields, b, c) => {
                  return (
                    <>
                      {fields.map((field) => (
                        <Form.Item
                          {...field}
                          label={attributesInitialValue[field.key].name}
                          name={[field.name, "value"]}
                          rules={[
                            {
                              required: true,
                              message: `Please enter ${
                                attributesInitialValue[field.key].name
                              }!`,
                            },
                          ]}
                        >
                          <Input />
                        </Form.Item>
                      ))}
                    </>
                  );
                }}
              </Form.List>

              <Form.Item wrapperCol={{ offset: 8, span: 16 }}>
                <Button
                  type="primary"
                  htmlType="submit"
                  disabled={store.issuingLoading}
                >
                  Submit offer
                </Button>
              </Form.Item>
            </Form>
          )}
        </div>
      </Modal.Body>
    );
  },
);

export default OfferCredentialDialog;
