import { action, makeAutoObservable, runInAction } from "mobx";
import api, { CreddefRecordDto, SchemaRecordDto } from "@dashboard/engine-api";
import { toast } from "react-toastify";
import displayError from "@dashboard/utils/displayError";

interface ProofUiAttribute {
  credentialDefinitionId: string | null;
  schemaLoading: boolean;
  schema: SchemaRecordDto | null;
  attributeName: string | null;
}

class RequestProofDialogStore {
  public initLoading = false;
  public connectionId: string | undefined = undefined;
  public credDefList: CreddefRecordDto[] = [];

  public rows: ProofUiAttribute[] = [];

  public issuingLoading = false;

  public get formIsValid() {
    if (!this.rows.length) {
      toast.error("Your need to create at least one row");
      return false;
    } else if (
      this.rows.some((p) => {
        return p.attributeName === null || p.credentialDefinitionId === null;
      })
    ) {
      toast.error("Please fill in all fields");
      return false;
    }
    return true;
  }

  private _schemaCache: {
    [id: string]: SchemaRecordDto;
  } = {};

  constructor() {
    makeAutoObservable(this);
  }

  public async init(connectionId: string | undefined) {
    runInAction(() => {
      this.initLoading = true;
      this.connectionId = connectionId;
    });
    try {
      const credDefList = await api.fetchCredentialDefinitions();
      runInAction(() => {
        this.addRow();
        this.credDefList = credDefList;
        this.initLoading = false;
      });
    } catch (e: unknown) {
      displayError(e);
      runInAction(() => (this.initLoading = false));
    }
  }

  public addRow = action(() => {
    this.rows.push({
      credentialDefinitionId: null,
      schemaLoading: false,
      schema: null,
      attributeName: null,
    });
  });

  public removeRow = action((index: number) => {
    this.rows.splice(index, 1);
  });

  public selectCredDef = async (index: number, credentialDefId: string) => {
    runInAction(() => {
      this.rows[index].credentialDefinitionId = credentialDefId;
      this.rows[index].schema = null;
      this.rows[index].schemaLoading = true;
      this.rows[index].attributeName = null;
    });
    const credDef = this.credDefList.find((p) => p.id === credentialDefId);
    const schemaId = credDef!.schemaId;

    try {
      if (this._schemaCache[schemaId]) {
        runInAction(() => {
          this.rows[index].schema = this._schemaCache[schemaId];
          this.rows[index].schemaLoading = false;
        });
        return;
      }

      const schema = await api.getSchemaById({
        id: credDef!.schemaId!,
      });
      runInAction(() => {
        this.rows[index].schema = schema;
        this.rows[index].schemaLoading = false;
        this._schemaCache[schemaId] = schema;
      });
    } catch (e: unknown) {
      displayError(e);
    }
  };

  public selectAttributeName = action(
    (index: number, attributeName: string) => {
      this.rows[index].attributeName = attributeName;
    },
  );

  public issueProofRequest = async () => {
    if (!this.formIsValid) return;

    runInAction(() => (this.issuingLoading = true));
    try {
      const proofRequest = await api.requestProof({
        connectionId: this.connectionId,
        attributes: this.rows.map((p) => ({
          attributeName: p.attributeName!,
          credentialDefinitionId: p.credentialDefinitionId!,
          schemaId: p.schema!.id!,
        })),
      });
      toast("Proof request created");
      runInAction(() => {
        this.issuingLoading = false;
      });
      return proofRequest;
    } catch (e: unknown) {
      displayError(e);
      runInAction(() => {
        this.issuingLoading = false;
      });
    }
    return null;
  };
}

export type { RequestProofDialogStore };

export default RequestProofDialogStore;
