import {
  Body,
  Controller,
  Get,
  Param,
  Post,
  Delete,
  UseFilters,
  Query,
} from "@nestjs/common";
import { AllExceptionsHandler } from "../exception.handler";
import { AttestationManagerClient } from "@ocm-engine/clients";
import {
  CreateCredentialDefinitionRequestDto,
  CreateSchemaRequestDto,
  CredentialFilterDto,
  OfferCredentialRequestDto,
  GatewayAcceptedResponseDto,
  CRED_DEF_CREATE,
  CRED_DEF_LIST,
  CRED_DEF_GET,
  CRED_SEND_OFFER,
  CRED_LIST,
  CRED_GET,
  CRED_OFFER_ACCEPT,
  CRED_OFFER_DECLINE,
  CRED_DELETE,
  MakeBasicMessageRequestDto,
  MessageFilterDto,
  MESSAGE_MAKE,
  MESSAGE_LIST,
  MESSAGE_DELETE,
  SCHEMA_CREATE,
  SCHEMA_GET,
  SCHEMA_LIST,
  IdReqDto,
  AcceptCredentialDto,
} from "@ocm-engine/dtos";
import { ApiBadRequestResponse, ApiOperation } from "@nestjs/swagger";
import IncludeOcmDtos from "../decorators/IncludeOcmDtos";
import OcmGatewayResponse from "../decorators/OcmGatewayResponse";
import OcmInternalServerErrorResponse from "../decorators/OcmInternalServerErrorResponse";

@UseFilters(AllExceptionsHandler)
@Controller("v1")
@IncludeOcmDtos()
export class AttestationController {
  constructor(private readonly amClient: AttestationManagerClient) {}

  @Post("/schemas")
  @OcmGatewayResponse()
  @OcmInternalServerErrorResponse("attestation manager")
  @ApiBadRequestResponse({
    description: "Validation error",
    content: {
      "application/json": {
        schema: {
          type: "object",
          properties: {
            statusCode: {
              type: "number",
              example: 400,
            },

            message: {
              type: "array",
              example: ["name must be a string", "name should not be empty"],
            },
            error: {
              type: "string",
              example: "Bad Request",
            },
          },
        },
      },
    },
  })
  @ApiOperation({
    summary: "Create schema",
    description:
      "Method will create schema. The id of the response will be matched when you receive event from the websocket",
    tags: ["Schema"],
  })
  async createSchema(
    @Body() schemaDto: CreateSchemaRequestDto,
  ): Promise<GatewayAcceptedResponseDto> {
    return this.amClient.sendPayload<CreateSchemaRequestDto>({
      pattern: "schemas",
      payload: {
        source: "/schema",
        data: schemaDto,
        type: SCHEMA_CREATE,
      },
    });
  }

  @Post("/schemas/get-by-id")
  @OcmGatewayResponse()
  @OcmInternalServerErrorResponse("attestation manager")
  @ApiBadRequestResponse({
    description: "Validation error",
    content: {
      "application/json": {
        schema: {
          type: "object",
          properties: {
            statusCode: {
              type: "number",
              example: 400,
            },

            message: {
              type: "array",
              example: ["id must be a string", "id should not be empty"],
            },
            error: {
              type: "string",
              example: "Bad Request",
            },
          },
        },
      },
    },
  })
  @ApiOperation({
    summary: "Get schema by id",
    description:
      "Method will fetch specific schema or throw exception. The id of the response will be matched when you receive event from the websocket",
    tags: ["Schema"],
  })
  async getSchemaById(
    @Body() dto: IdReqDto,
  ): Promise<GatewayAcceptedResponseDto> {
    return this.amClient.sendPayload<IdReqDto>({
      pattern: "schemas",
      payload: {
        source: "/schemas/get-by-id",
        data: dto,
        type: SCHEMA_GET,
      },
    });
  }

  @Get("/schemas")
  @OcmGatewayResponse()
  @OcmInternalServerErrorResponse("attestation manager")
  @ApiOperation({
    summary: "List all schemas",
    description:
      "Method will fetch all schemas. The id of the response will be matched when you receive event from the websocket",
    tags: ["Schema"],
  })
  async fetchSchemas(): Promise<GatewayAcceptedResponseDto> {
    return this.amClient.sendPayload<null>({
      pattern: "schemas",
      payload: {
        source: "/schema",
        data: null,
        type: SCHEMA_LIST,
      },
    });
  }

  @Get("/definitions")
  @OcmGatewayResponse()
  @OcmInternalServerErrorResponse("attestation manager")
  @ApiOperation({
    summary: "List all created credential definitions",
    description:
      "Method list credential definition no filters applied. The id of the response will be matched when you receive event from the websocket",
    tags: ["Credentials"],
  })
  async credentialDefinitions(): Promise<GatewayAcceptedResponseDto> {
    return this.amClient.sendPayload({
      pattern: "credentials",
      payload: {
        source: "/definitions",
        data: null,
        type: CRED_DEF_LIST,
      },
    });
  }

  @Post("/definitions/get-by-id")
  @OcmGatewayResponse()
  @OcmInternalServerErrorResponse("attestation manager")
  @ApiBadRequestResponse({
    description: "Validation error",
    content: {
      "application/json": {
        schema: {
          type: "object",
          properties: {
            statusCode: {
              type: "number",
              example: 400,
            },

            message: {
              type: "array",
              example: ["id must be a string", "id should not be empty"],
            },
            error: {
              type: "string",
              example: "Bad Request",
            },
          },
        },
      },
    },
  })
  @ApiOperation({
    summary: "Get credential definition by id",
    description:
      "Method will fetch specific credential definition or return null. The id of the response will be matched when you receive event from the websocket",
    tags: ["Credentials"],
  })
  async getCredentialDefinitionById(
    @Body() dto: IdReqDto,
  ): Promise<GatewayAcceptedResponseDto> {
    return this.amClient.sendPayload<IdReqDto>({
      pattern: "credentials",
      payload: {
        source: "/definitions/get-by-id",
        data: dto,
        type: CRED_DEF_GET,
      },
    });
  }

  @Post("/definitions")
  @OcmGatewayResponse()
  @OcmInternalServerErrorResponse("attestation manager")
  @ApiBadRequestResponse({
    description: "Validation error",
    content: {
      "application/json": {
        schema: {
          type: "object",
          properties: {
            statusCode: {
              type: "number",
              example: 400,
            },

            message: {
              type: "array",
              example: [
                "schemaId must be a string",
                "schemaId should not be empty",
                "tag must be a string",
                "tag should not be empty",
              ],
            },
            error: {
              type: "string",
              example: "Bad Request",
            },
          },
        },
      },
    },
  })
  @ApiOperation({
    summary: "Create credential definition",
    description:
      "Method create credential definition. The id of the response will be matched when you receive event from the websocket",
    tags: ["Credentials"],
  })
  async createCredentialDefinition(
    @Body() credentialDefinitionDto: CreateCredentialDefinitionRequestDto,
  ): Promise<GatewayAcceptedResponseDto> {
    return this.amClient.sendPayload({
      pattern: "credentials",
      payload: {
        source: "/definitions",
        data: credentialDefinitionDto,
        type: CRED_DEF_CREATE,
      },
    });
  }

  @Post("/credentials/offers")
  @OcmGatewayResponse()
  @OcmInternalServerErrorResponse("attestation manager")
  @ApiBadRequestResponse({
    description: "Validation error",
    content: {
      "application/json": {
        schema: {
          type: "object",
          properties: {
            statusCode: {
              type: "number",
              example: 400,
            },

            message: {
              type: "array",
              example: [
                "connectionId must be a string",
                "connectionId should not be empty",
                "credentialDefinitionId must be a string",
                "credentialDefinitionId should not be empty",
              ],
            },
            error: {
              type: "string",
              example: "Bad Request",
            },
          },
        },
      },
    },
  })
  @ApiOperation({
    summary: "Offer credential",
    description:
      "Method offer credential, it will create an offer and send it to specified receiver (connectionId). In case connectionId not provided, it will create connectionless credential offer. The id of the response will be matched when you receive event from the websocket",
    tags: ["Credentials"],
  })
  async offerCredential(
    @Body() dto: OfferCredentialRequestDto,
  ): Promise<GatewayAcceptedResponseDto> {
    return this.amClient.sendPayload({
      pattern: "credentials",
      payload: {
        source: "/credentials/offers",
        data: dto,
        type: CRED_SEND_OFFER,
      },
    });
  }

  @Get("/credentials")
  @OcmGatewayResponse()
  @OcmInternalServerErrorResponse("attestation manager")
  @ApiOperation({
    summary: "List all credential",
    description:
      "Method list credential definition based on provided filters. The id of the response will be matched when you receive event from the websocket",
    tags: ["Credentials"],
  })
  async fetchCredentials(
    @Query() credentialFilterDto: CredentialFilterDto,
  ): Promise<GatewayAcceptedResponseDto> {
    return this.amClient.sendPayload({
      pattern: "credentials",
      payload: {
        source: "/credentials",
        data: credentialFilterDto,
        type: CRED_LIST,
      },
    });
  }

  @Get("/credentials/:id")
  @OcmGatewayResponse()
  @ApiBadRequestResponse({
    description: "Validation error",
    content: {
      "application/json": {
        schema: {
          type: "object",
          properties: {
            statusCode: {
              type: "number",
              example: 400,
            },
            message: {
              type: "array",
              example: ["id must be a string", "id should not be empty"],
            },
            error: {
              type: "string",
              example: "Bad Request",
            },
          },
        },
      },
    },
  })
  @OcmInternalServerErrorResponse("attestation manager")
  @ApiOperation({
    summary: "Get credential by ID",
    description:
      "Method retrieves a specific credential based on the provided ID or returns null if not found. The ID of the response will be matched when you receive an event from the websocket",
    tags: ["Credentials"],
  })
  async getCredentialById(
    @Param("id") credentialId: string,
  ): Promise<GatewayAcceptedResponseDto> {
    const data = new IdReqDto();
    data.id = credentialId;

    return this.amClient.sendPayload<IdReqDto>({
      pattern: "credentials",
      payload: {
        source: "/credentials/:id",
        data,
        type: CRED_GET,
      },
    });
  }

  @Post("/credentials/offers/accept")
  @OcmGatewayResponse()
  @OcmInternalServerErrorResponse("attestation manager")
  @ApiOperation({
    summary: "Accept credential offers",
    description:
      "Method list accept credential offer. The id of the response will be matched when you receive event from the websocket",
    tags: ["Credentials Offers"],
  })
  async acceptCredential(
    @Body() dto: AcceptCredentialDto,
  ): Promise<GatewayAcceptedResponseDto> {
    return this.amClient.sendPayload({
      pattern: "credentials",
      payload: {
        source: "/credentials/offers/accept",
        data: dto,
        type: CRED_OFFER_ACCEPT,
      },
    });
  }

  @Post("/credentials/offers/:credential_record_id/decline")
  @OcmGatewayResponse()
  @OcmInternalServerErrorResponse("attestation manager")
  @ApiOperation({
    summary: "Decline credential offers",
    description:
      "Method to decline a credential offer. The id of the response will be matched when you receive an event from the websocket.",
    tags: ["Credentials Offers"],
  })
  async declineCredential(
    @Param("credential_record_id") credentialRecordId: string,
  ): Promise<GatewayAcceptedResponseDto> {
    const data = new IdReqDto();
    data.id = credentialRecordId;

    return this.amClient.sendPayload({
      pattern: "credentials",
      payload: {
        source: "/credentials/offers/:id/decline",
        data,
        type: CRED_OFFER_DECLINE,
      },
    });
  }

  @Delete("/credentials/:id")
  @OcmGatewayResponse()
  @OcmInternalServerErrorResponse("attestation manager")
  @ApiBadRequestResponse({
    description: "Validation error",
    content: {
      "application/json": {
        schema: {
          type: "object",
          properties: {
            statusCode: {
              type: "number",
              example: 400,
            },
            message: {
              type: "array",
              example: [
                "credentialId must be a string",
                "credentialId should not be empty",
              ],
            },
            error: {
              type: "string",
              example: "Bad Request",
            },
          },
        },
      },
    },
  })
  @ApiOperation({
    summary: "Delete a credential by ID",
    description:
      "Method deletes a specific credential based on provided ID. The id of the response will be matched when you receive an event from the websocket",
    tags: ["Credentials"],
  })
  async deleteCredentialById(
    @Param("id") credentialId: string,
  ): Promise<GatewayAcceptedResponseDto> {
    const data = new IdReqDto();
    data.id = credentialId;

    return this.amClient.sendPayload<IdReqDto>({
      pattern: "credentials",
      payload: {
        source: "/credentials/:id",
        data,
        type: CRED_DELETE,
      },
    });
  }

  @Post("/messages")
  @OcmGatewayResponse()
  @OcmInternalServerErrorResponse("attestation manager")
  @ApiBadRequestResponse({
    description: "Validation error",
    content: {
      "application/json": {
        schema: {
          type: "object",
          properties: {
            statusCode: {
              type: "number",
              example: 400,
            },

            message: {
              type: "array",
              example: [
                "connectionId must be a string",
                "connectionId should not be empty",
                "message must be a string",
                "message should not be empty",
              ],
            },
            error: {
              type: "string",
              example: "Bad Request",
            },
          },
        },
      },
    },
  })
  @ApiOperation({
    summary: "Send basic message",
    description:
      "Method will send basic message to a connection. The id of the response will be matched when you receive event from the websocket",
    tags: ["Credentials Offers"],
  })
  async sendMessage(
    @Body() message: MakeBasicMessageRequestDto,
  ): Promise<GatewayAcceptedResponseDto> {
    return this.amClient.sendPayload({
      pattern: "messages",
      payload: {
        source: "/messages",
        data: message,
        type: MESSAGE_MAKE,
      },
    });
  }

  @Get("/messages")
  @OcmGatewayResponse()
  @OcmInternalServerErrorResponse("attestation manager")
  @ApiBadRequestResponse({
    description: "Validation error",
    content: {
      "application/json": {
        schema: {
          type: "object",
          properties: {
            statusCode: {
              type: "number",
              example: 400,
            },
            message: {
              type: "array",
              example: [
                "message must be a string",
                "message should not be empty",
              ],
            },
            error: {
              type: "string",
              example: "Bad Request",
            },
          },
        },
      },
    },
  })
  @ApiOperation({
    summary: "Fetch messages based on filters",
    description:
      "Method will list and return messages based on the provided filters. The id of the response will be matched when you receive event from the websocket",
    tags: ["Messages"],
  })
  async fetchBasicMessages(
    @Query() filter: MessageFilterDto,
  ): Promise<GatewayAcceptedResponseDto> {
    return this.amClient.sendPayload<MessageFilterDto>({
      pattern: "messages",
      payload: {
        source: "/messages",
        data: filter,
        type: MESSAGE_LIST,
      },
    });
  }

  @Delete("/messages/:id")
  @OcmGatewayResponse()
  @OcmInternalServerErrorResponse("attestation manager")
  @ApiBadRequestResponse({
    description: "Validation error",
    content: {
      "application/json": {
        schema: {
          type: "object",
          properties: {
            statusCode: {
              type: "number",
              example: 400,
            },
            message: {
              type: "array",
              example: [
                "messageId must be a string",
                "messageId should not be empty",
              ],
            },
            error: {
              type: "string",
              example: "Bad Request",
            },
          },
        },
      },
    },
  })
  @ApiOperation({
    summary: "Delete message by ID",
    description:
      "Method will delete a specific message by ID. The id of the response will be matched when you receive an event from the websocket",
    tags: ["Messages"],
  })
  async deleteBasicMessage(
    @Param("id") messageId: string,
  ): Promise<GatewayAcceptedResponseDto> {
    const data = new IdReqDto();
    data.id = messageId;

    return this.amClient.sendPayload({
      pattern: "messages",
      payload: {
        source: "/messages/:id",
        data: data,
        type: MESSAGE_DELETE,
      },
    });
  }
}
