import {
  BadRequestException,
  Body,
  Controller,
  Delete,
  Get,
  Param,
  Post,
  UseFilters,
} from "@nestjs/common";
import { ConnectionManagerClient } from "@ocm-engine/clients";
import {
  CONNECTION_ACCEPT,
  CONNECTION_CREATE,
  CONNECTION_DELETE,
  CONNECTION_GET,
  CONNECTION_LIST,
  CreateInvitationResponseDto,
  GatewayAcceptedResponseDto,
  IdReqDto,
} from "@ocm-engine/dtos";
import { AllExceptionsHandler } from "../exception.handler";
import { ApiBadRequestResponse, ApiOperation } from "@nestjs/swagger";
import IncludeOcmDtos from "../decorators/IncludeOcmDtos";
import OcmGatewayResponse from "../decorators/OcmGatewayResponse";
import OcmInternalServerErrorResponse from "../decorators/OcmInternalServerErrorResponse";

@UseFilters(AllExceptionsHandler)
@Controller("v1")
@IncludeOcmDtos()
export class ConnectionController {
  constructor(private readonly cmClient: ConnectionManagerClient) {}

  @Post("/invitations")
  @OcmGatewayResponse()
  @OcmInternalServerErrorResponse("connection manager")
  @ApiOperation({
    summary: "Create invitation for connection",
    description:
      "Method will create invitation url. The id of the response will be matched when you receive event from the websocket",
    tags: ["Connections"],
  })
  async createInvitation(): Promise<GatewayAcceptedResponseDto> {
    return this.cmClient.sendPayload<null>({
      pattern: "connections",
      payload: {
        source: "/invitations",
        data: null,
        type: CONNECTION_CREATE,
      },
    });
  }

  @Post("/invitations/accept")
  @OcmGatewayResponse()
  @OcmInternalServerErrorResponse("connection manager")
  @ApiBadRequestResponse({
    description: "Validation error",
    content: {
      "application/json": {
        schema: {
          type: "object",
          properties: {
            statusCode: {
              type: "number",
              example: 400,
            },

            message: {
              type: "array",
              example: [
                "invitationUrl must be a string",
                "invitationUrl should not be empty",
              ],
            },
            error: {
              type: "string",
              example: "Bad Request",
            },
          },
        },
      },
    },
  })
  @ApiOperation({
    summary: "Accept invitation long and short urls for connection",
    description:
      "Method will accept long and short invitation urls and will return connection thought the websocket. The id of the response will be matched when you receive event from the websocket",
    tags: ["Connections"],
  })
  async acceptInvitation(
    @Body() createInvitationDto: CreateInvitationResponseDto,
  ): Promise<GatewayAcceptedResponseDto> {
    try {
      return this.cmClient.sendPayload<CreateInvitationResponseDto>({
        pattern: "connections",
        payload: {
          source: "/invitations/accept",
          data: createInvitationDto,
          type: CONNECTION_ACCEPT,
        },
      });
    } catch (e) {
      throw new BadRequestException(e);
    }
  }

  @Get("/connections")
  @OcmGatewayResponse()
  @OcmInternalServerErrorResponse("connection manager")
  @ApiOperation({
    summary: "List all connections",
    description:
      "The id of the response will be matched when you receive event from the websocket",
    tags: ["Connections"],
  })
  async fetchConnections(): Promise<GatewayAcceptedResponseDto> {
    return this.cmClient.sendPayload<null>({
      pattern: "connections",
      payload: {
        source: "/connections",
        data: null,
        type: CONNECTION_LIST,
      },
    });
  }

  @Get("/connections/:id")
  @OcmGatewayResponse()
  @OcmInternalServerErrorResponse("connection manager")
  @ApiOperation({
    summary: "Get connection by id",
    description:
      "The method will search for connection id, if not found null will be returned. The id of the response will be matched when you receive event from the websocket",
    tags: ["Connections"],
  })
  async getConnectionById(
    @Param("id") id: string,
  ): Promise<GatewayAcceptedResponseDto> {
    const request = new IdReqDto();
    request.id = id;

    return this.cmClient.sendPayload<IdReqDto>({
      pattern: "connections",
      payload: {
        source: "/connections/:id",
        data: request,
        type: CONNECTION_GET,
      },
    });
  }

  @Delete("/connections/:id")
  @OcmGatewayResponse()
  @OcmInternalServerErrorResponse("connection manager")
  @ApiOperation({
    summary: "Delete connection by id",
    description:
      "The method will attempt to delete a connection by its ID. The id of the response will be matched when you receive an event from the websocket.",
    tags: ["Connections"],
  })
  async deleteConnectionById(
    @Param("id") id: string,
  ): Promise<GatewayAcceptedResponseDto> {
    const request = new IdReqDto();
    request.id = id;

    return this.cmClient.sendPayload<IdReqDto>({
      pattern: "connections",
      payload: {
        source: "/connections/:id",
        data: request,
        type: CONNECTION_DELETE,
      },
    });
  }
}
