import { Injectable, Logger } from "@nestjs/common";
import { AgentService } from "../askar/agent.service";
import {
  AcceptProofDto,
  CloudEventDto,
  CONNECTION_ACCEPT,
  CONNECTION_CREATE,
  CONNECTION_GET,
  CONNECTION_LIST,
  CONNECTION_DELETE,
  CreateCredentialDefinitionRequestDto,
  CreateInvitationResponseDto,
  CreateSchemaRequestDto,
  CredentialFilterDto,
  CRED_DEF_CREATE,
  CRED_DEF_LIST,
  CRED_DEF_GET,
  CRED_SEND_OFFER,
  CRED_LIST,
  CRED_GET,
  CRED_OFFER_ACCEPT,
  CRED_OFFER_DECLINE,
  CRED_DELETE,
  OfferCredentialRequestDto,
  RequestProofDto,
  ProofFilterDto,
  PROOF_ACCEPT,
  PROOF_DECLINE,
  PROOF_REQUEST,
  PROOF_LIST,
  PROOF_GET,
  PROOF_DELETE,
  SCHEMA_CREATE,
  SCHEMA_GET,
  SCHEMA_LIST,
  MakeBasicMessageRequestDto,
  MessageFilterDto,
  MESSAGE_MAKE,
  MESSAGE_LIST,
  MESSAGE_DELETE,
  IdReqDto,
  AcceptCredentialDto,
  AcceptInvitationRequestDto,
} from "@ocm-engine/dtos";

@Injectable()
export class EventHandlerService {
  private readonly logger = new Logger(EventHandlerService.name);

  constructor(private readonly agentService: AgentService) {}

  async handle<T>(event: CloudEventDto<T>) {
    let data: unknown;
    let dto;
    switch (event.type) {
      case CONNECTION_CREATE:
        data = await this.agentService.createInvitation();
        break;

      case CONNECTION_ACCEPT:
        dto = event.data as AcceptInvitationRequestDto;
        data = await this.agentService.acceptInvitation(dto.invitationUrl);
        break;

      case CONNECTION_LIST:
        data = await this.agentService.fetchConnections();
        break;

      case CONNECTION_GET:
        dto = event.data as IdReqDto;
        data = await this.agentService.getConnectionById(dto.id);
        break;

      case CONNECTION_DELETE:
        dto = event.data as IdReqDto;
        data = await this.agentService.deleteConnectionById(dto.id);
        break;

      case SCHEMA_CREATE:
        dto = event.data as CreateSchemaRequestDto;
        data = await this.agentService.createSchema(dto);
        break;

      case SCHEMA_LIST:
        data = await this.agentService.fetchSchemas();
        break;

      case SCHEMA_GET:
        dto = event.data as IdReqDto;
        data = await this.agentService.getSchemaById(dto.id);
        break;

      case CRED_DEF_CREATE:
        data = await this.agentService.createCredentialDefinition(
          event.data as CreateCredentialDefinitionRequestDto,
        );
        break;

      case CRED_DEF_LIST:
        data = await this.agentService.fetchCredentialDefinitions();
        break;

      case CRED_DEF_GET:
        dto = event.data as IdReqDto;
        data = await this.agentService.getCredentialDefinitionById(dto.id);
        break;

      case CRED_SEND_OFFER:
        data = await this.agentService.offerCredential(
          event.data as OfferCredentialRequestDto,
        );
        break;

      case CRED_LIST:
        dto = event.data as CredentialFilterDto;
        data = await this.agentService.fetchCredentials(dto);
        break;

      case CRED_GET:
        dto = event.data as IdReqDto;
        data = await this.agentService.getCredentialById(dto.id);
        break;

      case CRED_OFFER_ACCEPT:
        data = await this.agentService.acceptCredential(
          event.data as AcceptCredentialDto,
        );
        break;

      case CRED_OFFER_DECLINE:
        dto = event.data as IdReqDto;
        data = await this.agentService.declineCredential(dto.id);
        break;

      case CRED_DELETE:
        dto = event.data as IdReqDto;
        data = await this.agentService.deleteCredentialById(dto.id);
        break;

      case PROOF_REQUEST:
        dto = event.data as RequestProofDto;
        data = await this.agentService.requestProof(dto);
        break;

      case PROOF_LIST:
        dto = event.data as ProofFilterDto;
        data = await this.agentService.fetchProofs(dto);
        break;

      case PROOF_GET:
        dto = event.data as IdReqDto;
        data = await this.agentService.getProofById(dto.id);
        break;

      case PROOF_ACCEPT:
        data = await this.agentService.acceptProof(
          event.data as AcceptProofDto,
        );
        break;

      case PROOF_DECLINE:
        dto = event.data as IdReqDto;
        data = await this.agentService.declineProofRequest(dto.id);
        break;

      case PROOF_DELETE:
        dto = event.data as IdReqDto;
        data = await this.agentService.deleteProofById(dto.id);
        break;

      case MESSAGE_MAKE:
        dto = event.data as MakeBasicMessageRequestDto;
        data = await this.agentService.sendMessage(dto);
        break;

      case MESSAGE_LIST:
        dto = event.data as MessageFilterDto;
        data = await this.agentService.fetchBasicMessages(dto);
        break;

      case MESSAGE_DELETE:
        dto = event.data as IdReqDto;
        data = await this.agentService.deleteMessageById(dto.id);
        break;
    }

    event.data = data as T;
    return event;
  }
}
