import {
  AcceptCredentialDto,
  AcceptProofDto,
  CreateCredentialDefinitionRequestDto,
  CreateInvitationRequestDto,
  CreateSchemaRequestDto,
  IdReqDto,
  MakeBasicMessageRequestDto,
  OfferCredentialRequestDto,
  RequestProofDto,
  RestControllerClient,
  Role,
  Roles,
  States,
  States2,
  States3,
  AcceptInvitationRequestDto,
} from "./frontend/agent_gen";

export * from "./frontend/agent_gen";

export interface Config {
  onUnauthorized?: () => void;
  wsUrl: string;
  httpUrl: string;
  getToken: () => Promise<string>;
}

let config: Config | null = null;

export const setConfig = (cfg: Config) => (config = cfg);

class ApiClient {
  private _ws!: WebSocket;
  private _rest!: RestControllerClient;

  public fetchInvitations = (
    states: States[] | undefined,
    roles: Roles[] | undefined,
  ) => this._rest.fetchInvitations(states, roles);
  public createInvitation = (body: CreateInvitationRequestDto) =>
    this._rest.createInvitation(body);
  public getInvitationById = (id: string) => this._rest.getInvitationById(id);
  public deleteInvitationById = (id: string) =>
    this._rest.deleteInvitationById(id);
  public acceptInvitation = (body: AcceptInvitationRequestDto) =>
    this._rest.acceptInvitation(body);
  public fetchConnections = () => this._rest.fetchConnections();
  public getConnectionById = (id: string) => this._rest.getConnectionById(id);
  public deleteConnectionById = (id: string) =>
    this._rest.deleteConnectionById(id);
  public getConnectionByOobId = (id: string) =>
    this._rest.getConnectionByOobId(id);
  public createSchema = (body: CreateSchemaRequestDto) =>
    this._rest.createSchema(body);
  public fetchSchemas = () => this._rest.fetchSchemas();
  public getSchemaById = (body: IdReqDto) => this._rest.getSchemaById(body);
  public fetchCredentialDefinitions = () =>
    this._rest.fetchCredentialDefinitions();
  public createCredentialDefinition = (
    body: CreateCredentialDefinitionRequestDto,
  ) => this._rest.createCredentialDefinition(body);
  public getCredentialDefinitionById = (body: IdReqDto) =>
    this._rest.getCredentialDefinitionById(body);
  public offerCredential = (body: OfferCredentialRequestDto) =>
    this._rest.offerCredential(body);
  public fetchCredentials = (
    states: States2[] | undefined,
    connectionId: string | undefined,
  ) => this._rest.fetchCredentials(states, connectionId);
  public getCredentialById = (id: string) => this._rest.getCredentialById(id);
  public deleteCredentialById = (id: string) =>
    this._rest.deleteCredentialById(id);
  public getCredentialFormatDataById = (id: string) =>
    this._rest.getCredentialFormatDataById(id);
  public acceptCredential = (body: AcceptCredentialDto) =>
    this._rest.acceptCredential(body);
  public declineCredential = (credential_record_id: string) =>
    this._rest.declineCredential(credential_record_id);
  public sendMessage = (body: MakeBasicMessageRequestDto) =>
    this._rest.sendMessage(body);
  public fetchBasicMessages = (
    role: Role | undefined,
    connectionId: string | undefined,
  ) => this._rest.fetchBasicMessages(role, connectionId);
  public deleteBasicMessage = (id: string) => this._rest.deleteBasicMessage(id);
  public fetchProofs = (
    states: States3[] | undefined,
    connectionId: string | undefined,
  ) => this._rest.fetchProofs(states, connectionId);
  public getProofById = (proof_record_id: string) =>
    this._rest.getProofById(proof_record_id);
  public deleteProofById = (proof_record_id: string) =>
    this._rest.deleteProofById(proof_record_id);
  public getProofFormatDataById = (proof_record_id: string) =>
    this._rest.getProofFormatDataById(proof_record_id);
  public proofAcceptanceWait = (proof_record_id: string) =>
    this._rest.proofAcceptanceWait(proof_record_id);
  public requestProof = (body: RequestProofDto) =>
    this._rest.requestProof(body);
  public acceptProof = (body: AcceptProofDto) => this._rest.acceptProof(body);
  public declineProofRequest = (proof_record_id: string) =>
    this._rest.declineProofRequest(proof_record_id);
  public resolveDid = (body: IdReqDto) => this._rest.resolveDid(body);
  public getCreatedDids = () => this._rest.getCreatedDids();

  public initialize() {
    const _config = config as Config;

    this._rest = new RestControllerClient(_config.httpUrl, {
      // eslint-disable-next-line @typescript-eslint/ban-ts-comment
      // @ts-ignore
      fetch: this._fetch.bind(this),
    });
    this._rest.jsonParseReviver = this.jsonParseReviver;

    // this._listen();
  }

  public async getToken(): Promise<string> {
    if (!config) {
      throw new Error("ApiClient: Please call setConfig before calling api");
    }

    return config.getToken();
  }

  private async _fetch(url: RequestInfo, init?: RequestInit): Promise<unknown> {
    if (!config) {
      throw new Error("ApiClient: Please call setConfig before calling api");
    }

    init = Object.assign({}, init, {
      headers: {
        ...(init?.headers || {}),
        Authorization: `Bearer ${await config.getToken()}`,
      },
    });

    const response = await fetch(url, init);

    if (response.status === 401) {
      if (config.onUnauthorized) {
        config.onUnauthorized();
      }
    }

    return response;
  }

  public jsonParseReviver = (key: string, value: unknown): unknown => {
    if (typeof value === "string") {
      const a =
        /^(\d{4})-(\d{2})-(\d{2})T(\d{2}):(\d{2}):(\d{2}(?:\.\d*)?)Z?$/.exec(
          value,
        );
      if (a) {
        return new Date(value);
      }
    }
    return value;
  };

  private async _listen() {
    console.log("Opening connection to " + config!.wsUrl);
    const ws = new WebSocket(config!.wsUrl);
    this._ws = ws;
    // Connection opened
    ws.addEventListener("open", (event) => {
      console.log("Connected to the WebSocket");
    });

    // Listen for messages
    ws.addEventListener("message", (event) => {
      console.log("Message from server: ", event.data);
      const data = JSON.parse(event.data, this.jsonParseReviver);

      console.warn("Message not processed. No handler");
    });

    // Connection closed
    ws.addEventListener("close", (event) => {
      if (event.wasClean) {
        console.log(
          `Closed cleanly, code=${event.code}, reason=${event.reason}`,
        );
      } else {
        console.log("Connection died");
      }
    });

    // Connection errors
    ws.addEventListener("error", (error) => {
      console.error(`WebSocket Error: ${error}`);
    });
  }
}

export default new ApiClient();
